@@ -1,4 +1,3 @@
-import typing
 from pkg_resources import get_distribution, DistributionNotFound
 
 try:
@@ -23,6 +22,8 @@ from datetime import datetime, date
 import stat
 import tempfile
 from pathlib import Path
+import time
+import sqlite3
 import sys
 import typing
 from typing import (Any, Callable, Iterator, List, NamedTuple, Optional, Tuple,
@@ -31,22 +32,17 @@ import dataclasses
 import warnings
 
 
-import sqlalchemy # type: ignore
+import appdirs  # type: ignore[import]
+
+import sqlalchemy  # type: ignore[import]
 from sqlalchemy import Column, Table, event
 
 
-if sys.version_info[1] < 7:
-    from .compat import fromisoformat
-else:
-    fromisoformat = datetime.fromisoformat
-
-
 from .compat import fix_sqlalchemy_StatementError_str
 try:
     fix_sqlalchemy_StatementError_str()
 except Exception as e:
-    # todo warn or something??
-    pass
+    logging.exception(e)
 
 
 # in case of changes in the way cachew stores data, this should be changed to discard old caches
@@ -64,8 +60,7 @@ class settings:
     '''
     ENABLE: bool = True
 
-    # switch to appdirs and using user cache dir?
-    DEFAULT_CACHEW_DIR: PathIsh = Path(tempfile.gettempdir()) / 'cachew'
+    DEFAULT_CACHEW_DIR: PathIsh = Path(appdirs.user_cache_dir('cachew'))
 
     '''
     Set to true if you want to fail early. Otherwise falls back to non-cached version
@@ -104,7 +99,10 @@ class IsoDateTime(sqlalchemy.TypeDecorator):
             return iso
         else:
             if isinstance(tz, pytz.BaseTzInfo):
-                return iso + ' ' + tz.zone
+                zone = tz.zone
+                # should be present: https://github.com/python/typeshed/blame/968fd6d01d23470e0c8368e7ee7c43f54aaedc0e/stubs/pytz/pytz/tzinfo.pyi#L6
+                assert zone is not None, tz
+                return iso + ' ' + zone
             else:
                 return iso
 
@@ -112,7 +110,7 @@ class IsoDateTime(sqlalchemy.TypeDecorator):
         if value is None:
             return None
         spl = value.split(' ')
-        dt = fromisoformat(spl[0])
+        dt = datetime.fromisoformat(spl[0])
         if len(spl) <= 1:
             return dt
         zone = spl[1]
@@ -332,23 +330,12 @@ def strip_generic(tp):
     >>> strip_generic(str)
     <class 'str'>
     """
-    if sys.version_info[1] < 7:
-        # pylint: disable=no-member
-        if isinstance(tp, typing.GenericMeta):
-            return tp.__extra__ # type: ignore
-    else:
-        GA = getattr(typing, '_GenericAlias') # ugh, can't make both mypy and pylint happy here?
-        if isinstance(tp, GA):
-            return tp.__origin__
+    GA = getattr(typing, '_GenericAlias') # ugh, can't make both mypy and pylint happy here?
+    if isinstance(tp, GA):
+        return tp.__origin__
     return tp
 
 
-# release mode friendly assert
-def kassert(x: bool) -> None:
-    if x is False:
-        raise AssertionError
-
-
 NT = TypeVar('NT')
 # sadly, bound=NamedTuple is not working yet in mypy
 # https://github.com/python/mypy/issues/685
@@ -427,7 +414,11 @@ class NTBinder(NamedTuple):
                 fields = ()
                 span = 1
             else:
-                fields = tuple(NTBinder.make(tp=ann, name=fname) for fname, ann in tp.__annotations__.items())
+                annotations = getattr(tp, '__annotations__', None)
+                # https://www.python.org/dev/peps/pep-3107/#accessing-function-annotations
+                if annotations is None or annotations == {}:
+                    raise CachewException(f"{tp}: doesn't look like a supported type to cache. See https://github.com/karlicoss/cachew#features for the list of supported types.")
+                fields = tuple(NTBinder.make(tp=ann, name=fname) for fname, ann in annotations.items())
                 span = sum(f.span for f in fields) + (1 if optional else 0)
         return NTBinder(
             name=name,
@@ -474,7 +465,7 @@ class NTBinder(NamedTuple):
                 is_none = obj is None
                 yield is_none
             else:
-                is_none = False; kassert(obj is not None)  # TODO hmm, that last assert is not very symmetric...
+                is_none = False; assert obj is not None  # TODO hmm, that last assert is not very symmetric...
 
             if is_none:
                 for _ in range(self.span - 1):
@@ -496,7 +487,8 @@ class NTBinder(NamedTuple):
                 r
                 for r in uf._from_row(row_iter) if r is not None
             ]
-            kassert(len(union_params) == 1); return union_params[0]
+            assert len(union_params) == 1, union_params
+            return union_params[0]
         else:
             if self.optional:
                 is_none = next(row_iter)
@@ -505,7 +497,8 @@ class NTBinder(NamedTuple):
 
             if is_none:
                 for _ in range(self.span - 1):
-                    x = next(row_iter); kassert(x is None)  # huh. assert is kinda opposite of producing value
+                    x = next(row_iter)
+                    assert x is None, x  # huh. assert is kinda opposite of producing value
                 return None
             else:
                 return self.type_(*(
@@ -557,11 +550,7 @@ class DbHelper:
         # NOTE: timeout is necessary so we don't lose time waiting during recursive calls
         # by default, it's several seconds? you'd see 'test_recursive' test performance degrade
 
-        import sqlite3
-        import time
-
-        from sqlalchemy.engine import Engine # type: ignore
-        @event.listens_for(Engine, 'connect')
+        @event.listens_for(self.engine, 'connect')
         def set_sqlite_pragma(dbapi_connection, connection_record):
             # without wal, concurrent reading/writing is not gonna work
 
@@ -572,7 +561,8 @@ class DbHelper:
                     break
                 except sqlite3.OperationalError as oe:
                     if 'database is locked' not in str(oe):
-                        raise oe
+                        # ugh, pretty annoying that exception doesn't include database path for some reason
+                        raise CachewException(f'Error while setting WAL on {db_path}') from oe
                 time.sleep(0.1)
 
 
@@ -598,9 +588,10 @@ class DbHelper:
         self.table_hash = Table('hash', self.meta, Column('value', sqlalchemy.String))
 
         self.binder = NTBinder.make(tp=cls)
-        self.table_data = Table('table', self.meta, *self.binder.columns)
-
-        # TODO FIXME database/tables need to be created atomically?
+        # actual cache
+        self.table_cache     = Table('cache'    , self.meta, *self.binder.columns)
+        # temporary table, we use it to insert and then (atomically?) rename to the above table at the very end
+        self.table_cache_tmp = Table('cache_tmp', self.meta, *self.binder.columns)
 
     def __enter__(self):
         return self
@@ -627,7 +618,6 @@ def mtime_hash(path: Path, *args, **kwargs) -> SourceHash:
 Failure = str
 
 
-# pylint: disable=too-many-return-statements
 def infer_type(func) -> Union[Failure, Type[Any]]:
     """
     >>> from typing import Collection, NamedTuple
@@ -676,7 +666,7 @@ def infer_type(func) -> Union[Failure, Type[Any]]:
     if is_union(arg):
         return arg # meh?
     if not is_dataclassish(arg):
-        return bail(f"{arg} is not NamedTuple")
+        return bail(f"{arg} is not NamedTuple/dataclass")
     return arg
 
 
@@ -702,14 +692,13 @@ def cachew_error(e: Exception) -> None:
     else:
         logger = get_logger()
         # todo add func name?
-        logger.error("Error while setting up cache, falling back to non-cached version")
+        logger.error("cachew: error while setting up cache, falling back to non-cached version")
         logger.exception(e)
 
 
 use_default_path = cast(Path, object())
 
 @doublewrap
-# pylint: disable=too-many-arguments
 def cachew(
         func=None,
         cache_path: Optional[PathProvider]=use_default_path,
@@ -724,6 +713,7 @@ def cachew(
         # you can use 'test_many' to experiment
         # - too small values (e.g. 10)  are slower than 100 (presumably, too many sql statements)
         # - too large values (e.g. 10K) are slightly slower as well (not sure why?)
+        synthetic_key: Optional[str]=None,
         **kwargs,
 ):
     r"""
@@ -776,7 +766,7 @@ def cachew(
     cn = cname(func)
     # todo not very nice that ENABLE check is scattered across two places
     if not settings.ENABLE or cache_path is None:
-        logger.info('[%s]: cache disabled', cn)
+        logger.info('[%s]: cache explicitly disabled (settings.ENABLE is False or cache_path is None)', cn)
         return func
 
     if cache_path is use_default_path:
@@ -786,7 +776,7 @@ def cachew(
     # TODO fuzz infer_type, should never crash?
     inferred = infer_type(func)
     if isinstance(inferred, Failure):
-        msg = f"failed to infer cache type: {inferred}"
+        msg = f"failed to infer cache type: {inferred}. See https://github.com/karlicoss/cachew#features for the list of supported types."
         if cls is None:
             ex = CachewException(msg)
             cachew_error(ex)
@@ -804,13 +794,14 @@ def cachew(
                 # TODO not sure if should be more serious error...
 
     ctx = Context(
-        func      =func,
-        cache_path=cache_path,
-        force_file=force_file,
-        cls       =cls,
-        depends_on=depends_on,
-        logger    =logger,
-        chunk_by  =chunk_by,
+        func         =func,
+        cache_path   =cache_path,
+        force_file   =force_file,
+        cls          =cls,
+        depends_on   =depends_on,
+        logger       =logger,
+        chunk_by     =chunk_by,
+        synthetic_key=synthetic_key,
     )
 
     # hack to avoid extra stack frame (see test_recursive, test_deep-recursive)
@@ -827,16 +818,23 @@ def cname(func: Callable) -> str:
     return f'{mod}:{func.__qualname__}'
 
 
-class Context(NamedTuple):
-    func      : Callable
-    cache_path: PathProvider
-    force_file: bool
-    cls       : Type
-    depends_on: HashFunction
-    logger    : logging.Logger
-    chunk_by  : int
+_CACHEW_CACHED       = 'cachew_cached'  # TODO add to docs
+_SYNTHETIC_KEY       = 'synthetic_key'
+_SYNTHETIC_KEY_VALUE = 'synthetic_key_value'
+_DEPENDENCIES        = 'dependencies'
 
-    def composite_hash(self, *args, **kwargs) -> SourceHash:
+
+class Context(NamedTuple):
+    func         : Callable
+    cache_path   : PathProvider
+    force_file   : bool
+    cls          : Type
+    depends_on   : HashFunction
+    logger       : logging.Logger
+    chunk_by     : int
+    synthetic_key: Optional[str]
+
+    def composite_hash(self, *args, **kwargs) -> Dict[str, Any]:
         fsig = inspect.signature(self.func)
         # defaults wouldn't be passed in kwargs, but they can be an implicit dependency (especially inbetween program runs)
         defaults = {
@@ -853,8 +851,20 @@ class Context(NamedTuple):
         }
         kwargs = {**defaults, **kwargs}
         binder = NTBinder.make(tp=self.cls)
-        schema = str(binder.columns) # TODO not super nice, but works fine for now
-        return f'cachew: {CACHEW_VERSION}, schema: {schema}, dependencies: {self.depends_on(*args, **kwargs)}'
+        schema = str(binder.columns) # todo not super nice, but works fine for now
+        hash_parts = {
+            'cachew'      : CACHEW_VERSION,
+            'schema'      : schema,
+            _DEPENDENCIES : str(self.depends_on(*args, **kwargs)),
+        }
+        synthetic_key = self.synthetic_key
+        if synthetic_key is not None:
+            hash_parts[_SYNTHETIC_KEY      ] = synthetic_key
+            hash_parts[_SYNTHETIC_KEY_VALUE] = kwargs[synthetic_key]
+            # FIXME assert it's in kwargs in the first place?
+            # FIXME support positional args too? maybe extract the name from signature somehow? dunno
+            # need to test it
+        return hash_parts
 
 
 def cachew_wrapper(
@@ -863,17 +873,18 @@ def cachew_wrapper(
         **kwargs,
 ):
     C = _cachew_context
-    func       = C.func
-    cache_path = C.cache_path
-    force_file = C.force_file
-    cls        = C.cls
-    depend_on  = C.depends_on
-    logger     = C.logger
-    chunk_by   = C.chunk_by
+    func          = C.func
+    cache_path    = C.cache_path
+    force_file    = C.force_file
+    cls           = C.cls
+    depends_on    = C.depends_on
+    logger        = C.logger
+    chunk_by      = C.chunk_by
+    synthetic_key = C.synthetic_key
 
     cn = cname(func)
     if not settings.ENABLE:
-        logger.info('[%s]: cache disabled', cn)
+        logger.info('[%s]: cache explicitly disabled (settings.ENABLE is False)', cn)
         yield from func(*args, **kwargs)
         return
 
@@ -887,7 +898,7 @@ def cachew_wrapper(
         if callable(cache_path):
             pp = cache_path(*args, **kwargs) # type: ignore
             if pp is None:
-                logger.info('[%s]: cache disabled', cn)
+                logger.info('[%s]: cache explicitly disabled (cache_path is None)', cn)
                 yield from func(*args, **kwargs)
                 return
             else:
@@ -915,65 +926,129 @@ def cachew_wrapper(
 
         logger.debug('using %s for db cache', dbp)
 
-        h = C.composite_hash(*args, **kwargs); kassert(h is not None)  # just in case
-        logger.debug('new hash: %s', h)
+        new_hash_d = C.composite_hash(*args, **kwargs)
+        new_hash = json.dumps(new_hash_d)
+        logger.debug('new hash: %s', new_hash)
 
         with DbHelper(dbp, cls) as db, \
              db.connection.begin():
-            # NOTE: defferred transaction
+            # NOTE: deferred transaction
             conn = db.connection
             binder = db.binder
-            values_table = db.table_data
+            table_cache     = db.table_cache
+            table_cache_tmp = db.table_cache_tmp
 
             # first, try to do as much as possible read-only, benefiting from deferred transaction
             try:
                 # not sure if there is a better way...
-                prev_hashes = conn.execute(db.table_hash.select()).fetchall()
+                old_hashes = conn.execute(db.table_hash.select()).fetchall()
             except sqlalchemy.exc.OperationalError as e:
                 # meh. not sure if this is a good way to handle this..
                 if 'no such table: hash' in str(e):
-                    prev_hashes = []
+                    old_hashes = []
                 else:
                     raise e
 
-            kassert(len(prev_hashes) <= 1)  # shouldn't happen
-            prev_hash: Optional[SourceHash]
-            if len(prev_hashes) == 0:
-                prev_hash = None
+            assert len(old_hashes) <= 1, old_hashes  # shouldn't happen
+            old_hash: Optional[SourceHash]
+            if len(old_hashes) == 0:
+                old_hash = None
             else:
-                prev_hash = prev_hashes[0][0]  # returns a tuple...
+                old_hash = old_hashes[0][0]  # returns a tuple...
 
-            logger.debug('old hash: %s', prev_hash)
+            logger.debug('old hash: %s', old_hash)
 
-            if h == prev_hash:
-                logger.debug('hash matched: loading from cache')
-                rows = conn.execute(values_table.select())
+
+            def cached_items():
+                rows = conn.execute(table_cache.select())
                 for row in rows:
                     yield binder.from_row(row)
+
+            if new_hash == old_hash:
+                logger.debug('hash matched: loading from cache')
+                yield from cached_items()
                 return
 
             logger.debug('hash mismatch: computing data and writing to db')
 
+            if synthetic_key is not None:
+                # attempt to use existing cache if possible, as a 'prefix'
+
+                old_hash_d: Dict[str, Any] = {}
+                if old_hash is not None:
+                    try:
+                        old_hash_d = json.loads(old_hash)
+                    except json.JSONDecodeError:
+                        # possible if we used old cachew version (<=0.8.1), hash wasn't json
+                        pass
+
+                hash_diffs = {
+                    k: new_hash_d.get(k) == old_hash_d.get(k)
+                    for k in (*new_hash_d.keys(), *old_hash_d.keys())
+                    # the only 'allowed' differences for hash, otherwise need to recompute (e.g. if schema changed)
+                    if k not in {_SYNTHETIC_KEY_VALUE, _DEPENDENCIES}
+                }
+                cache_compatible = all(hash_diffs.values())
+                if cache_compatible:
+                    def missing_keys(cached: List[str], wanted: List[str]) -> Optional[List[str]]:
+                        # FIXME assert both cached and wanted are sorted? since we rely on it
+                        # if not, then the user could use some custom key for caching (e.g. normalise filenames etc)
+                        # although in this case passing it into the function wouldn't make sense?
+
+                        if len(cached) == 0:
+                            # no point trying to reuse anything, cache should be empty?
+                            return None
+                        if len(wanted) == 0:
+                            # similar, no way to reuse cache
+                            return None
+                        if cached[0] != wanted[0]:
+                            # there is no common prefix, so no way to reuse cache really
+                            return None
+                        last_cached = cached[-1]
+                        # ok, now actually figure out which items are missing
+                        for i, k in enumerate(wanted):
+                            if k > last_cached:
+                                # ok, rest of items are missing
+                                return wanted[i:]
+                        # otherwise too many things are cached, and we seem to wante less
+                        return None
+
+                    new_values: List[str] = new_hash_d[_SYNTHETIC_KEY_VALUE]
+                    old_values: List[str] = old_hash_d[_SYNTHETIC_KEY_VALUE]
+                    missing = missing_keys(cached=old_values, wanted=new_values)
+                    if missing is not None:
+                        # can reuse cache
+                        kwargs[_CACHEW_CACHED] = cached_items()
+                        kwargs[synthetic_key] = missing
+
+
             # NOTE on recursive calls
             # somewhat magically, they should work as expected with no extra database inserts?
             # the top level call 'wins' the write transaction and once it's gathered all data, will write it
             # the 'intermediate' level calls fail to get it and will pass data through
-            # the cached 'bottom' level is read only and will be yielded withotu a write transaction
-
+            # the cached 'bottom' level is read only and will be yielded without a write transaction
             try:
-                # first write statement will upgrade transaction to write transaction which might fail due to concurrency
+                # first 'write' statement will upgrade transaction to write transaction which might fail due to concurrency
                 # see https://www.sqlite.org/lang_transaction.html
-                # note I guess, because of 'checkfirst', only the last create is actually guaranteed to upgrade the transaction to write one
-                # drop and create to incorporate schema changes
+                # NOTE: because of 'checkfirst=True', only the last .create will guarantee the transaction upgrade to write transaction
                 db.table_hash.create(conn, checkfirst=True)
-                values_table.drop(conn, checkfirst=True)
-                values_table.create(conn)
+
+                # 'table' used to be old 'cache' table name, so we just delete it regardless
+                # otherwise it might overinfalte the cache db with stale values
+                conn.execute(f'DROP TABLE IF EXISTS `table`')
+
+                # NOTE: we have to use .drop and then .create (e.g. instead of some sort of replace)
+                # since it's possible to have schema changes inbetween calls
+                # checkfirst=True because it might be the first time we're using cache
+                table_cache_tmp.drop(conn, checkfirst=True)
+                table_cache_tmp.create(conn)
             except sqlalchemy.exc.OperationalError as e:
-                if e.code == 'e3q8':
-                    # database is locked; someone else must be have won the write lock
+                if e.code == 'e3q8' and 'database is locked' in str(e):
+                    # someone else must be have won the write lock
                     # not much we can do here
                     # NOTE: important to close early, otherwise we might hold onto too many file descriptors during yielding
                     # see test_deep_recursive
+                    # (normally connection is closed in DbHelper.__exit__)
                     conn.close()
                     yield from func(*args, **kwargs)
                     return
@@ -982,13 +1057,20 @@ def cachew_wrapper(
             # at this point we're guaranteed to have an exclusive write transaction
 
             datas = func(*args, **kwargs)
+            column_names = [c.name for c in table_cache_tmp.columns]
+            insert_into_table_cache_tmp = table_cache_tmp.insert()
 
             chunk: List[Any] = []
-            def flush():
+            def flush() -> None:
                 nonlocal chunk
                 if len(chunk) > 0:
-                    # pylint: disable=no-value-for-parameter
-                    conn.execute(values_table.insert().values(chunk))
+                    # TODO hmm, it really doesn't work unless you zip into a dict first
+                    # maybe should return dicts from binder instead then?
+                    chunk_dict = [
+                        dict(zip(column_names, row))
+                        for row in chunk
+                    ]
+                    conn.execute(insert_into_table_cache_tmp, chunk_dict)
                     chunk = []
 
             for d in datas:
@@ -1003,12 +1085,20 @@ def cachew_wrapper(
                     flush()
             flush()
 
-            # TODO insert and replace instead?
-
+            # delete hash first, so if we are interrupted somewhere, it mismatches next time and everything is recomputed
             # pylint: disable=no-value-for-parameter
             conn.execute(db.table_hash.delete())
+
+            # checkfirst is necessary since it might not have existed in the first place
+            # e.g. first time we use cache
+            table_cache.drop(conn, checkfirst=True)
+
+            # meh https://docs.sqlalchemy.org/en/14/faq/metadata_schema.html#does-sqlalchemy-support-alter-table-create-view-create-trigger-schema-upgrade-functionality
+            # also seems like sqlalchemy doesn't have any primitives to escape table names.. sigh
+            conn.execute(f"ALTER TABLE `{table_cache_tmp.name}` RENAME TO `{table_cache.name}`")
+
             # pylint: disable=no-value-for-parameter
-            conn.execute(db.table_hash.insert().values([{'value': h}]))
+            conn.execute(db.table_hash.insert().values([{'value': new_hash}]))
     except Exception as e:
         # sigh... see test_early_exit_shutdown...
         if early_exit and 'Cannot operate on a closed database' in str(e):
