@@ -1,7 +1,8 @@
 from datetime import datetime, date, timezone
 import inspect
-from itertools import islice
+from itertools import islice, chain
 import logging
+import os
 from pathlib import Path
 from random import Random
 from subprocess import check_call, run, PIPE
@@ -9,9 +10,9 @@ import string
 import sys
 import time
 import timeit
-from typing import NamedTuple, Iterator, Optional, List, Set, Tuple, cast, Iterable, Dict, Any, Union
+from typing import NamedTuple, Iterator, Optional, List, Set, Tuple, cast, Iterable, Dict, Any, Union, Sequence
 
-from more_itertools import one
+from more_itertools import one, ilen, last, unique_everseen
 
 import pytz
 import pytest  # type: ignore
@@ -23,7 +24,8 @@ logger = get_logger()
 
 @pytest.fixture(autouse=True)
 def throw_on_errors():
-    # a more reasonable default for tests
+    # NOTE: in tests we always throw on errors, it's a more reasonable default for testing.
+    # we still check defensive behaviour in test_defensive
     settings.THROW_ON_ERROR = True
     yield
 
@@ -40,7 +42,7 @@ def restore_settings():
 
 @pytest.fixture
 def tdir(tmp_path):
-    # TODO just use tmp_path instead?
+    # todo just use tmp_path instead?
     yield Path(tmp_path)
 
 
@@ -160,7 +162,43 @@ def inner(_it, _timer{init}):
     assert t < 2.0, 'should be pretty much instantaneous'
 
 
-def test_cache_path(tdir):
+def test_error(tmp_path: Path) -> None:
+    '''
+    Test behaviour when the first time cache is initialized it ends up with an error
+    '''
+    cache_file = tmp_path / 'cache'
+    assert not cache_file.exists(), cache_file  # just precondition
+
+    should_raise = True
+
+    @cachew(cache_file, force_file=True)
+    def fun() -> Iterator[str]:
+        yield 'string1'
+        if should_raise:
+            raise RuntimeError('oops')
+        yield 'string2'
+
+    with pytest.raises(RuntimeError, match='oops'):
+        list(fun())
+
+    # vvv this would be nice but might be tricky because of the way sqlite works (i.e. wal mode creates a file)
+    # assert not cache_file.exists(), cache_file
+
+    # perhaps doesn't hurt either way as long this vvv works properly
+    # shouldn't cache anything and crach again
+    with pytest.raises(RuntimeError, match='oops'):
+        list(fun())
+
+    should_raise = False
+    assert list(fun()) == ['string1', 'string2']
+
+
+def test_cache_path(tmp_path: Path) -> None:
+    '''
+    Tests various ways of specifying cache path
+    '''
+    tdir = tmp_path
+
     calls = 0
     def orig() -> Iterable[int]:
         nonlocal calls
@@ -216,36 +254,60 @@ def test_cache_path(tdir):
     # on the other hand, wouldn't want to delete some user file by accident
 
 
+class UGood(NamedTuple):
+    x: int
+class UBad:
+    pass
+
+def test_unsupported_class(tmp_path: Path) -> None:
+    with pytest.raises(CachewException, match='.*failed to infer cache type.*'):
+        @cachew(cache_path=tmp_path)
+        def fun() -> List[UBad]:
+            return [UBad()]
+
+    # now something a bit nastier
+    # TODO hmm, should really throw at the definition time... but can fix later I suppose
+    @cachew(cache_path=tmp_path)
+    def fun2() -> Iterable[Union[UGood, UBad]]:
+        yield UGood(x=1)
+        yield UBad()
+        yield UGood(x=2)
+
+    with pytest.raises(CachewException, match=".*doesn't look like a supported type.*"):
+        list(fun2())
+
+
 class TE2(NamedTuple):
     value: int
     uuu: UUU
     value2: int
 
 
-# TODO also profile datetimes?
-def test_many(tdir):
-    COUNT = 100000
-    logger = get_logger()
-
-    src = tdir / 'source'
+# you can run one specific test (e.g. to profile) by passing it as -k to pytest
+# e.g. -k 'test_many[500000-False]'
+@pytest.mark.parametrize('count,on_ci', [
+    (100000, True ),
+    (500000, False),
+])
+def test_many(count: int, on_ci: bool, tmp_path: Path) -> None:
+    if 'CI' in os.environ and not on_ci:
+        pytest.skip("test would be too slow on CI anyway, only meant to run manually")
+    # should be a parametrized test perhaps
+    src = tmp_path / 'source'
     src.touch()
 
-    @cachew(cache_path=lambda path: tdir / (path.name + '.cache'), cls=TE2)
-    def _iter_data(_: Path):
-        for i in range(COUNT):
+    @cachew(cache_path=lambda path: tmp_path / (path.name + '.cache'))
+    def _iter_data(_: Path) -> Iterator[TE2]:
+        for i in range(count):
+            # TODO also profile datetimes?
             yield TE2(value=i, uuu=UUU(xx=i, yy=i), value2=i)
 
-    def iter_data():
+    def iter_data() -> Iterator[TE2]:
         return _iter_data(src)
 
-    def ilen(it):
-        ll = 0
-        for _ in it:
-            ll += 1
-        return ll
-    assert ilen(iter_data()) == COUNT
-    assert ilen(iter_data()) == COUNT
-    logger.debug('done')
+    assert ilen(iter_data()) == count  # initial
+    assert ilen(iter_data()) == count  # hitting cache
+    assert last(iter_data()) == TE2(value=count - 1, uuu=UUU(xx=count - 1, yy=count - 1), value2=count - 1)
 
     # serializing to db
     # in-memory: 16 seconds
@@ -504,13 +566,13 @@ def test_stats(tdir):
     print(f"Cache db size for {N} entries: estimated size {one * N // 1024} Kb, actual size {cache_file.stat().st_size // 1024} Kb;")
 
 
-def test_dataclass(tdir):
+def test_dataclass(tmp_path: Path) -> None:
     from dataclasses import dataclass
     @dataclass
     class Test:
         field: int
 
-    @cachew(tdir)
+    @cachew(tmp_path)
     def get_dataclasses() -> Iterator[Test]:
         yield from [Test(field=i) for i in range(5)]
 
@@ -618,7 +680,7 @@ def test_primitive(tmp_path: Path):
 class O(NamedTuple):
     x: int
 
-def test_default_arguments(tmp_path: Path):
+def test_default_arguments(tmp_path: Path) -> None:
     class HackHash:
         def __init__(self, x: int) -> None:
             self.x = x
@@ -675,7 +737,7 @@ def test_default_arguments(tmp_path: Path):
 class U(NamedTuple):
     x: Union[str, O]
 
-def test_union(tmp_path: Path):
+def test_union(tmp_path: Path) -> None:
     @cachew(tmp_path)
     def fun() -> Iterator[U]:
         yield U('hi')
@@ -685,6 +747,21 @@ def test_union(tmp_path: Path):
     assert list(fun()) == [U('hi'), U(O(123))]
 
 
+def test_union_with_dataclass(tmp_path: Path) -> None:
+    from dataclasses import dataclass
+    # NOTE: empty dataclass doesn't have __annotations__ ??? not sure if need to handle it...
+    @dataclass
+    class DD:
+        x: int
+
+    @cachew(tmp_path)
+    def fun() -> Iterator[Union[int, DD]]:
+        yield 123
+        yield DD(456)
+
+    assert list(fun()) == [123, DD(456)]
+
+
 def _concurrent_helper(cache_path: Path, count: int, sleep_s=0.1):
     from time import sleep
     @cachew(cache_path)
@@ -707,14 +784,14 @@ def fuzz_cachew_impl():
     patch = '''\
 @@ -740,6 +740,11 @@
 
-             logger.debug('old hash: %s', prev_hash)
+             logger.debug('old hash: %s', old_hash)
 
 +            from random import random
 +            rs = random() * 2
 +            print("sleeping for: ", rs)
 +            from time import sleep; sleep(rs)
 +
-             if h == prev_hash:
+             if new_hash == old_hash:
                  logger.debug('hash matched: loading from cache')
                  rows = conn.execute(values_table.select())
 '''
@@ -804,7 +881,7 @@ def test_defensive(restore_settings):
     assert list(fun()) == [123]
     assert list(fun()) == [123]
 
-    from cachew.compat import nullcontext
+    from contextlib import nullcontext
     for throw in [True, False]:
         ctx = pytest.raises(Exception) if throw else nullcontext()
         settings.THROW_ON_ERROR = throw
@@ -1045,39 +1122,41 @@ def test_disabled(tmp_path: Path):
         assert calls == 3
 
 
-def test_early_exit(tmp_path: Path):
-    cf = 0
+def test_early_exit(tmp_path: Path) -> None:
+    # cachew works on iterators and we'd prefer not to cache if the iterator hasn't been exhausted
+    calls_f = 0
     @cachew(tmp_path) #  / 'fun', force_file=True)
     def f() -> Iterator[int]:
         yield from range(20)
-        nonlocal cf
-        cf += 1
+        nonlocal calls_f
+        calls_f += 1
 
-    cg = 0
+    calls_g = 0
     @cachew(tmp_path) # / 'fun', force_file=True)
     def g() -> Iterator[int]:
         yield from f()
-        nonlocal cg
-        cg += 1
+        nonlocal calls_g
+        calls_g += 1
 
+    # only consume 10/20 items
     assert len(list(islice(g(), 0, 10))) == 10
-    assert cf == 0 # hasn't finished
-    assert cg == 0 # hasn't finished
+    # precondition
+    assert calls_f == 0  # f hasn't been fully exhausted
+    assert calls_g == 0  # g hasn't been fully exhausted
 
     # todo not sure if need to check that db is empty?
-   
     assert len(list(g())) == 20
-    assert cf == 1
-    assert cg == 1
+    assert calls_f == 1
+    assert calls_g == 1
 
     # should be cached now
     assert len(list(g())) == 20
-    assert cf == 1
-    assert cg == 1
+    assert calls_f == 1
+    assert calls_g == 1
 
 
 # see https://github.com/sqlalchemy/sqlalchemy/issues/5522#issuecomment-705156746
-def test_early_exit_shutdown(tmp_path: Path):
+def test_early_exit_shutdown(tmp_path: Path) -> None:
     # don't ask... otherwise the exception doesn't appear :shrug:
     import_hack = '''
 from sqlalchemy import Column
@@ -1104,3 +1183,95 @@ print("FINISHED")
     r = run(['python3', '-c', prog], cwd=tmp_path, stderr=PIPE, stdout=PIPE, check=True)
     assert r.stdout.strip() == b'FINISHED'
     assert b'Traceback' not in r.stderr
+
+
+# tests both modes side by side to demonstrate the difference
+@pytest.mark.parametrize('use_synthetic', ['False', 'True'])
+def test_synthetic_keyset(tmp_path: Path, use_synthetic: bool) -> None:
+    # just to keep track of which data we had to compute from scratch
+    _recomputed: List[str] = []
+
+    # assume key i is responsible for numbers i and i-1
+    # in reality this could be some slow function we'd like to avoid calling if its results is already cached
+    # e.g. the key would typically be a filename (e.g. isoformat timestamp)
+    # and the returned values could be the results of an export over the month prior to the timestamp, or something like that
+    # see https://beepb00p.xyz/exports.html#synthetic for more on the motivation
+    def compute(key: str) -> Iterator[str]:
+        _recomputed.append(key)
+        n = int(key)
+        yield str(n - 1)
+        yield str(n)
+
+
+    # should result in 01 + 12 + 45                     == 01245
+    keys125         = ['1', '2', '5'                    ]
+    # should result in 01 + 12 + 45 + 56 + 67           == 0124567
+    keys12567       = ['1', '2', '5', '6', '7'          ]
+    # should result in 01 + 12 + 45 + 56      + 78 + 89 == 012456789
+    keys125689      = ['1', '2', '5', '6',      '8', '9']
+    # should result in           45 + 56      + 78 + 89 ==    456789
+    keys5689        = [          '5', '6',      '8', '9']
+
+
+    def recomputed() -> List[str]:
+        r = list(_recomputed)
+        _recomputed.clear()
+        return r
+
+
+    ## 'cachew_cached' will just be [] if synthetic key is not used, so no impact on data
+    @cachew(tmp_path, synthetic_key=('keys' if use_synthetic else None))
+    def fun_aux(keys: Sequence[str], *, cachew_cached: Iterable[str] = []) -> Iterator[str]:
+        yield from unique_everseen(chain(
+            cachew_cached,
+            *(compute(key) for key in keys),
+        ))
+
+    def fun(keys: Sequence[str]) -> Set[str]:
+        return set(fun_aux(keys=keys))
+    ##
+
+    assert fun(keys125) == set('01' '12' '45')
+    assert recomputed() == keys125
+    assert fun(keys125) == set('01' '12' '45')
+    assert recomputed() == []  # should be cached
+
+    assert fun(keys12567) == set('01' '12' '45' '56' '67')
+    if use_synthetic:
+        # 1, 2 and 5 should be already cached from the previous call
+        assert recomputed() == ['6', '7']
+    else:
+        # but without synthetic key this would cause everything to recompute
+        assert recomputed() == keys12567
+    assert fun(keys12567) == set('01' '12' '45' '56' '67')
+    assert recomputed() == []  # should be cached
+
+    assert fun(keys125689) == set('01' '12' '45' '56' '78' '89')
+    if use_synthetic:
+        # similarly, 1 2 5 6 7 are cached from the previous cacll
+        assert recomputed() == ['8', '9']
+    else:
+        # and we need to call against all keys otherwise
+        assert recomputed() == keys125689
+    assert fun(keys125689) == set('01' '12' '45' '56' '78' '89')
+    assert recomputed() == []  # should be cached
+
+    assert fun(keys5689) == set('45' '56' '78' '89')
+    # now the prefix has changed, so if we returned cached items it might return too much
+    # so have to recompute everything
+    assert recomputed() == keys5689
+    assert fun(keys5689) == set('45' '56' '78' '89')
+    assert recomputed() == []  # should be cached
+
+    # TODO maybe call combined function? so it could return total result and last cached?
+    # TODO another option is:
+    # the function yields all cached stuff first
+    # then the user yields stuff from new
+    # and then external function does merging
+    # TODO test with kwargs hash?...
+    # TODO try without and with simultaneously?
+    # TODO check what happens when errors happen?
+    # FIXME check what happens if we switch between modes? (synthetic/non-synthetic)
+    # FIXME make sure this thing works if len(keys) > chunk size?
+    # TODO check what happens when we forget to set 'cachew_cached' argument
+    # TODO check what happens when keys are not str but e.g. Path
